# -*- coding: utf-8 -*-
# Copyright (c) Spyder Project Contributors
#
# Licensed under the terms of the MIT License
# (see LICENSE.txt for details)

# Standard library imports
import os.path as osp

# Third-party imports
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QMessageBox, QVBoxLayout

# Spyder imports
from spyder.api.widgets.main_widget import PluginMainWidget
from spyder.config.gui import is_dark_interface
from spyder.utils.switcher import shorten_paths

# Local imports
from spyder_notebook.utils.localization import _
from spyder_notebook.utils.servermanager import ServerManager
from spyder_notebook.widgets.notebooktabwidget import NotebookTabWidget
from spyder_notebook.widgets.serverinfo import ServerInfoDialog


class NotebookMainWidgetToolButtons:
    NewNotebook = 'New notebook'


class NotebookMainWidgetActions:
    NewNotebook = 'New notebook'
    SaveAs = 'Save as'
    Open = 'Open'
    OpenConsole = 'Open console'
    ServerInfo = 'Server info'
    ClearRecentNotebooks = 'Clear recent notebooks'
    RecentNotebook = 'Recent notebook'


class NotebookMainWidgetMenus:
    RecentNotebooks = 'Recent notebooks'


class NotebookMainWidgetOptionsMenuSections:
    Open = 'Open'
    Other = 'Other'


class NotebookMainWidgetRecentNotebooksMenuSections:
    Notebooks = 'Notebooks'
    Clear = 'Clear'


class NotebookMainWidget(PluginMainWidget):

    sig_open_console_requested = Signal(str, str)
    """
    Request to open an IPython console associated to a notebook.

    Parameters
    -----------
    kernel_id: str
        Id of the kernel to open a console for.
    tab_name: str
        Tab name to set for the created console.
    """

    def __init__(self, name, plugin, parent):
        """Widget constructor."""
        super().__init__(name, plugin, parent)

        self.recent_notebooks = self.get_conf('recent_notebooks', default=[])
        self.server_manager = ServerManager(self.dark_theme)

        # Tab widget
        self.tabwidget = NotebookTabWidget(
            self,
            self.server_manager,
            dark_theme=self.dark_theme
        )
        self.tabwidget.currentChanged.connect(self.refresh_plugin)

        # Widget layout
        layout = QVBoxLayout()
        layout.addWidget(self.tabwidget)
        self.setLayout(layout)

        # Connect to switcher
        self.switcher = plugin.main.switcher
        self.switcher.sig_mode_selected.connect(self.handle_switcher_modes)
        self.switcher.sig_item_selected.connect(
            self.handle_switcher_selection)

    # ---- PluginMainWidget API
    # ------------------------------------------------------------------------
    def get_focus_widget(self):
        """Return the widget to give focus to."""
        client = self.tabwidget.currentWidget()
        if client is not None:
            return client.notebookwidget

    def get_title(self):
        """Define the title of the widget."""
        return _('Notebook')

    def setup(self):
        """Perform the setup of plugin's main menu and signals."""
        # Corner widgets
        new_notebook_toolbar_action = self.create_toolbutton(
            NotebookMainWidgetToolButtons.NewNotebook,
            icon=self.create_icon('options_more'),
            tip=_('Open a new notebook'),
            triggered=self.create_new_client
        )

        self.add_corner_widget(
            NotebookMainWidgetToolButtons.NewNotebook,
            new_notebook_toolbar_action)

        # Menu actions
        new_notebook_action = self.create_action(
            NotebookMainWidgetActions.NewNotebook,
            text=_("New notebook"),
            icon=self.create_icon('filenew'),
            triggered=self.create_new_client
        )
        open_notebook_action = self.create_action(
            NotebookMainWidgetActions.Open,
            text=_("Open..."),
            icon=self.create_icon('fileopen'),
            triggered=self.open_notebook
        )
        self.save_as_action = self.create_action(
            NotebookMainWidgetActions.SaveAs,
            text=_("Save as..."),
            icon=self.create_icon('filesaveas'),
            triggered=self.save_as
        )
        self.open_console_action = self.create_action(
            NotebookMainWidgetActions.OpenConsole,
            text=_("Open console"),
            icon=self.create_icon('ipython_console'),
            triggered=self.open_console
        )
        self.server_info_action = self.create_action(
            NotebookMainWidgetActions.ServerInfo,
            text=_('Server info...'),
            icon=self.create_icon('log'),
            triggered=self.view_servers
        )
        self.clear_recent_notebooks_action = self.create_action(
            NotebookMainWidgetActions.ClearRecentNotebooks,
            text=_("Clear this list"),
            triggered=self.clear_recent_notebooks
        )

        # Submenu
        self.recent_notebooks_menu = self.create_menu(
            NotebookMainWidgetMenus.RecentNotebooks,
            _("Open recent")
        )

        # Options menu
        options_menu = self.get_options_menu()
        for item in [new_notebook_action, open_notebook_action,
                     self.recent_notebooks_menu]:
            self.add_item_to_menu(
                item,
                menu=options_menu,
                section=NotebookMainWidgetOptionsMenuSections.Open,
            )

        for item in [self.save_as_action, self.open_console_action,
                     self.server_info_action]:
            self.add_item_to_menu(
                item,
                menu=options_menu,
                section=NotebookMainWidgetOptionsMenuSections.Other,
            )

    def update_actions(self):
        """Update actions of the options menu."""
        try:
            client = self.tabwidget.currentWidget()
        except AttributeError:  # tabwidget is not yet constructed
            client = None

        if client and not self.tabwidget.is_welcome_client(client):
            self.save_as_action.setEnabled(True)
            self.open_console_action.setEnabled(True)
        else:
            self.save_as_action.setEnabled(False)
            self.open_console_action.setEnabled(False)

    def on_close(self):
        """
        Perform actions before parent main window is closed.

        This function closes all tabs, shuts down all notebook server and
        stores the file names of all opened notebooks that are not temporary
        and all notebooks in the 'Open recent' menu in Spyder's config.
        """
        opened_notebooks = []
        for client_index in range(self.tabwidget.count()):
            client = self.tabwidget.widget(client_index)
            if (not self.tabwidget.is_welcome_client(client)
                    and not self.tabwidget.is_newly_created(client)):
                opened_notebooks.append(client.filename)
            client.close()

        self.set_conf('recent_notebooks', self.recent_notebooks)
        self.set_conf('opened_notebooks', opened_notebooks)
        self.server_manager.shutdown_all_servers()

    # ---- Public API
    # ------------------------------------------------------------------------
    @property
    def dark_theme(self):
        """Whether to use the dark theme for notebooks (bool)."""
        theme_config = self.get_conf('theme', default='same as spyder')
        if theme_config == 'same as spyder':
            return is_dark_interface()
        elif theme_config == 'dark':
            return True
        elif theme_config == 'light':
            return False
        else:
            raise RuntimeError('theme config corrupted, value = {}'
                               .format(theme_config))

    def refresh_plugin(self):
        """Refresh tabwidget."""
        nb = None
        if self.tabwidget.count():
            client = self.tabwidget.currentWidget()
            nb = client.notebookwidget
            nb.setFocus()
        else:
            nb = None

    def update_recent_notebooks_menu(self):
        """Update the recent notebooks menu actions."""
        self.recent_notebooks_menu.clear_actions()
        if self.recent_notebooks:
            for notebook in self.recent_notebooks:
                # Create notebook action
                name = notebook
                action = self.create_action(
                    NotebookMainWidgetActions.RecentNotebook,
                    text=name,
                    icon=self.create_icon('notebook'),
                    register_action=False,
                    triggered=lambda v, path=notebook:
                        self.create_new_client(filename=path)
                )

                # Add action to menu
                self.add_item_to_menu(
                    action,
                    menu=self.recent_notebooks_menu,
                    section=NotebookMainWidgetRecentNotebooksMenuSections.Notebooks,
                )

        self.add_item_to_menu(
            self.clear_recent_notebooks_action,
            menu=self.recent_notebooks_menu,
            section=NotebookMainWidgetRecentNotebooksMenuSections.Clear,
        )

        if self.recent_notebooks:
            self.clear_recent_notebooks_action.setEnabled(True)
        else:
            self.clear_recent_notebooks_action.setEnabled(False)

    def open_previous_session(self):
        """Open notebooks left open in the previous session."""
        filenames = self.get_conf('opened_notebooks')
        if filenames:
            self.open_notebook(filenames)
        else:
            self.tabwidget.maybe_create_welcome_client()
            self.create_new_client()
            self.tabwidget.setCurrentIndex(0)  # bring welcome tab to top

    def open_notebook(self, filenames=None):
        """Open a notebook from file."""
        filenames = self.tabwidget.open_notebook(filenames)
        for filename in filenames:
            self.add_to_recent(filename)
        self.update_recent_notebooks_menu()

    def create_new_client(self, filename=None):
        """Create a new notebook or load a pre-existing one."""
        client = self.tabwidget.create_new_client(filename)
        if not self.tabwidget.is_newly_created(client):
            self.add_to_recent(filename)
            self.update_recent_notebooks_menu()

    def add_to_recent(self, notebook):
        """
        Add an entry to recent notebooks.

        We only maintain the list of the 20 most recent notebooks.
        """
        if notebook not in self.recent_notebooks:
            self.recent_notebooks.insert(0, notebook)
            self.recent_notebooks = self.recent_notebooks[:20]

    def save_as(self):
        """Save current notebook to different file."""
        self.tabwidget.save_as()

    def open_console(self, client=None):
        """Open an IPython console for the given client or the current one."""
        if not client:
            client = self.tabwidget.currentWidget()

        kernel_id = client.get_kernel_id()
        if not kernel_id:
            QMessageBox.critical(
                self,
                _('Error opening console'),
                _('There is no kernel associated to this notebook.')
            )
            return

        self.sig_open_console_requested.emit(
            kernel_id,
            client.get_short_name()
        )

    def view_servers(self):
        """Display server info."""
        dialog = ServerInfoDialog(self.server_manager.servers, parent=self)
        dialog.show()

    def clear_recent_notebooks(self):
        """Clear the list of recent notebooks."""
        self.recent_notebooks = []
        self.update_recent_notebooks_menu()

    def handle_switcher_modes(self, mode):
        """
        Populate switcher with opened notebooks.

        List the file names of the opened notebooks with their directories in
        the switcher. Only handle file mode, where `mode` is empty string.
        """
        if mode != '':
            return

        clients = [self.tabwidget.widget(i)
                   for i in range(self.tabwidget.count())]
        paths = [client.get_filename() for client in clients]
        is_unsaved = [False for client in clients]
        short_paths = shorten_paths(paths, is_unsaved)
        icon = self.create_icon('notebook')
        section = self.get_title()

        for path, short_path, client in zip(paths, short_paths, clients):
            title = osp.basename(path)
            description = osp.dirname(path)
            if len(path) > 75:
                description = short_path
            is_last_item = (client == clients[-1])

            self.switcher.add_item(
                title=title,
                description=description,
                icon=icon,
                section=section,
                data=client,
                last_item=is_last_item
            )

    def handle_switcher_selection(self, item, mode, search_text):
        """
        Handle user selecting item in switcher.

        If the selected item is not in the section of the switcher that
        corresponds to this plugin, then ignore it. Otherwise, switch to
        selected item in notebook plugin and hide the switcher.
        """
        if item.get_section() != self.get_title():
            return

        client = item.get_data()
        index = self.tabwidget.indexOf(client)
        self.tabwidget.setCurrentIndex(index)
        self._plugin.switch_to_plugin()
        self.switcher.hide()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder notebook default configuration."""

CONF_SECTION = 'notebook'

CONF_DEFAULTS = [
    (
        CONF_SECTION,
        {
            'recent_notebooks': [],    # Items in "Open recent" menu
            'opened_notebooks': [],    # Notebooks to open at start
            'theme': 'same as spyder'  # Notebook theme (light/dark)
        }
    )
]

# IMPORTANT NOTES:
# 1. If you want to *change* the default value of a current option, you need to
#    do a MINOR update in config version, e.g. from 1.0.0 to 1.1.0
# 2. If you want to *remove* options that are no longer needed in our codebase,
#    or if you want to *rename* options, then you need to do a MAJOR update in
#    version, e.g. from 1.0.0 to 2.0.0
# 3. You don't need to touch this value if you're just adding a new option
CONF_VERSION = '0.1.0'



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License

"""File implementing ServerManager."""

# Standard library imports
import datetime
import enum
import json
import logging
import os.path as osp
import sys

# Qt imports
from qtpy.QtCore import QObject, QProcess, QProcessEnvironment, QTimer, Signal

# Third-party imports
from jupyter_core.paths import jupyter_runtime_dir
from notebook import notebookapp

# Spyder imports
from spyder.config.base import DEV, get_home_dir, get_module_path


# Kernel specification to use in notebook server
KERNELSPEC = 'spyder.plugins.ipythonconsole.utils.kernelspec.SpyderKernelSpec'

# Delay we wait to check whether server is up (in ms)
CHECK_SERVER_UP_DELAY = 250

# Delay before we give up on server starting (in s)
SERVER_TIMEOUT_DELAY = 30

logger = logging.getLogger(__name__)


class ServerState(enum.Enum):
    """State of a server process."""

    STARTING = 1
    RUNNING = 2
    FINISHED = 3
    ERROR = 4
    TIMED_OUT = 5


class ServerProcess:
    """
    Process executing a notebook server.

    This is a data class.
    """

    def __init__(self, process, notebook_dir, interpreter, starttime=None,
                 state=ServerState.STARTING, server_info=None, output=''):
        """
        Construct a ServerProcess.

        Parameters
        ----------
        process : QProcess
            The process described by this instance.
        notebook_dir : str
            Directory from which the server can render notebooks.
        interpreter : str
            File name of Python interpreter used to render notebooks.
        starttime : datetime or None, optional
            Time at which the process was started. The default is None,
            meaning that the current time should be used.
        state : ServerState, optional
            State of the server process. The default is ServerState.STARTING.
        server_info : dict or None, optional
            If set, this is a dict with information given by the server in
            a JSON file in jupyter_runtime_dir(). It has keys like 'url' and
            'token'. The default is None.
        output : str
            Output of the server process from stdout and stderr. The default
            is ''.
        """
        self.process = process
        self.notebook_dir = notebook_dir
        self.interpreter = interpreter
        self.starttime = starttime or datetime.datetime.now()
        self.state = state
        self.server_info = server_info
        self.output = output


class ServerManager(QObject):
    """
    Manager for notebook servers.

    A Jupyter notebook server will only render notebooks under a certain
    directory, so we may need several servers. This class manages all these
    servers.

    Attributes
    ----------
    dark_theme : bool
        Whether notebooks should be rendered using the dark theme.
    servers : list of ServerProcess
        List of servers managed by this object.
    """

    # A server has started and is now accepting requests
    sig_server_started = Signal(ServerProcess)

    # We tried to start a server but it took too long to start up
    sig_server_timed_out = Signal(ServerProcess)

    # We tried to start a server but an error occurred
    sig_server_errored = Signal(ServerProcess)

    def __init__(self, dark_theme=False):
        """
        Construct a ServerManager.

        Parameters
        ----------
        dark_theme : bool, optional
            Whether notebooks should be rendered using the dark theme.
            The default is False.
        """
        super().__init__()
        self.dark_theme = dark_theme
        self.servers = []

    def get_server(self, filename, interpreter, start=True):
        """
        Return server which can render a notebook or potentially start one.

        Return the server info of a server managed by this object which can
        render the notebook with the given file name and which uses the given
        interpreter. If no such server exists and `start` is True, then start
        up a server asynchronously (unless a suitable server is already in the
        process of starting up).

        Parameters
        ----------
        filename : str
            File name of notebook which is to be rendered.
        interpreter : str
            File name of Python interpreter to be used.
        start : bool, optional
            Whether to start up a server if none exists. The default is True.

        Returns
        -------
        dict or None
            A dictionary describing the server which can render the notebook,
            or None if no such server exists.
        """
        filename = osp.abspath(filename)
        for server in self.servers:
            if (filename.startswith(server.notebook_dir)
                    and interpreter == server.interpreter):
                if server.state == ServerState.RUNNING:
                    return server.server_info
                elif server.state == ServerState.STARTING:
                    logger.debug('Waiting for server for %s to start up',
                                 server.notebook_dir)
                    return None
        if start:
            self.start_server(filename, interpreter)
        return None

    def start_server(self, filename, interpreter):
        """
        Start a notebook server asynchronously.

        Start a server which can render the given notebook and return
        immediately. Assume the server uses the given interpreter. The manager
        will check periodically whether the server is accepting requests and
        emit `sig_server_started` or `sig_server_timed_out` when appropriate.

        Parameters
        ----------
        filename : str
            File name of notebook to be rendered by the server.
        interpreter : str
            File name of Python interpreter to be used.
        """
        home_dir = get_home_dir()
        if filename.startswith(home_dir):
            nbdir = home_dir
        else:
            nbdir = osp.dirname(filename)

        logger.debug('Starting new notebook server for %s', nbdir)
        process = QProcess(None)
        serverscript = osp.join(osp.dirname(__file__), '../server/main.py')
        serverscript = osp.normpath(serverscript)
        arguments = [serverscript, '--no-browser',
                     '--notebook-dir={}'.format(nbdir),
                     '--NotebookApp.password=',
                     '--KernelSpecManager.kernel_spec_class={}'.format(
                           KERNELSPEC)]
        if self.dark_theme:
            arguments.append('--dark')
        logger.debug('Arguments: %s', repr(arguments))

        if DEV:
            env = QProcessEnvironment.systemEnvironment()
            env.insert('PYTHONPATH', osp.dirname(get_module_path('spyder')))
            process.setProcessEnvironment(env)

        server_process = ServerProcess(
            process, notebook_dir=nbdir, interpreter=interpreter)
        process.setProcessChannelMode(QProcess.MergedChannels)
        process.readyReadStandardOutput.connect(
            lambda: self.read_server_output(server_process))
        process.errorOccurred.connect(
            lambda error: self.handle_error(server_process, error))
        process.finished.connect(
            lambda code, status:
                self.handle_finished(server_process, code, status))

        process.start(sys.executable, arguments)
        self.servers.append(server_process)

        self._check_server_started(server_process)

    def _check_server_started(self, server_process):
        """
        Check whether a notebook server has started up.

        If the server state is no longer in the "starting" state (probably
        because an error occurred or the server exited prematurely) then
        do nothing.

        Otherwise, look for a json file in the Jupyter runtime dir to check
        whether the notebook server has started up. If so, then emit
        `sig_server_started` and fill the server info with the contents of the
        json file. If not, then schedule another check after a short delay
        (as set in CHECK_SERVER_UP_DELAY) unless the server is taken too long
        (as specified by SERVER_TIMEOUT_DELAY). In the latter case, emit
        sig_server_timed_out`.

        Parameters
        ----------
        server_process : ServerProcess
            The server process to be checked.
        """
        if server_process.state != ServerState.STARTING:
            return

        pid = server_process.process.processId()
        runtime_dir = jupyter_runtime_dir()
        filename = osp.join(runtime_dir, 'nbserver-{}.json'.format(pid))

        try:
            with open(filename, encoding='utf-8') as f:
                server_info = json.load(f)
        except OSError:  # E.g., file does not exist
            delay = datetime.datetime.now() - server_process.starttime
            if delay > datetime.timedelta(seconds=SERVER_TIMEOUT_DELAY):
                logger.debug('Notebook server for %s timed out',
                             server_process.notebook_dir)
                server_process.state = ServerState.TIMED_OUT
                self.sig_server_timed_out.emit(server_process)
            else:
                QTimer.singleShot(
                    CHECK_SERVER_UP_DELAY,
                    lambda: self._check_server_started(server_process))
            return None

        logger.debug('Server for %s started', server_process.notebook_dir)
        server_process.state = ServerState.RUNNING
        server_process.server_info = server_info
        self.sig_server_started.emit(server_process)

    def shutdown_all_servers(self):
        """Shutdown all running servers."""
        for server in self.servers:
            if server.state == ServerState.RUNNING:
                logger.debug('Shutting down notebook server for %s',
                             server.notebook_dir)
                server.process.errorOccurred.disconnect()
                server.process.finished.disconnect()
                notebookapp.shutdown_server(server.server_info)
                server.state = ServerState.FINISHED

    def read_server_output(self, server_process):
        """
        Read the output of the notebook server process.

        This function is connected to the QProcess.readyReadStandardOutput
        signal. It reads the contents of the standard output channel of the
        server process and stores it in `server_process.output`. The standard
        error channel is merged into the standard output channel.

        Parameters
        ----------
        server_process : ServerProcess
            The server process whose output is to be read.
        """
        byte_array = server_process.process.readAllStandardOutput()
        output = byte_array.data().decode(errors='backslashreplace')
        server_process.output += output

    def handle_error(self, server_process, error):
        """
        Handle errors that occurred in the notebook server process.

        This function is connected to the QProcess.errorOccurred signal.
        It changes the state of the process and emits `sig_server_errored`.

        Parameters
        ----------
        server_process : ServerProcess
            The server process that encountered an error
        error : QProcess.ProcessError
            Error as determined by Qt.
        """
        logger.debug('Server for %s encountered error %s',
                     server_process.notebook_dir, str(error))
        server_process.state = ServerState.ERROR
        self.sig_server_errored.emit(server_process)

    def handle_finished(self, server_process, code, status):
        """
        Handle signal that notebook server process has finished.

        This function is connected to the QProcess.finished signal.
        It changes the state of the process.

        Parameters
        ----------
        server_process : ServerProcess
            The server process that encountered an error.
        code : int
            Exit code as reported by Qt.
        status : QProcess.ExitStatus
            Exit status as reported by Qt.
        """
        logger.debug('Server for %s finished with code = %d, status = %s',
                     server_process.notebook_dir, code, str(status))
        server_process.state = ServerState.FINISHED



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""File implementing NotebookTabWidget."""

# Standard library imports
import logging
import os
import os.path as osp
import sys

# Qt imports
from qtpy.compat import getopenfilenames, getsavefilename
from qtpy.QtCore import QEventLoop, QTimer
from qtpy.QtWidgets import QMessageBox

# Third-party imports
import nbformat

# Spyder imports
from spyder.api.config.mixins import SpyderConfigurationAccessor
from spyder.utils.misc import get_python_executable
from spyder.utils.programs import get_temp_dir
from spyder.widgets.tabs import Tabs

# Local imports
from spyder_notebook.utils.localization import _
from spyder_notebook.widgets.client import NotebookClient


# Directory in which new notebooks are created
NOTEBOOK_TMPDIR = osp.join(get_temp_dir(), 'notebooks')

# Path to HTML file with welcome message
PACKAGE_PATH = osp.join(osp.dirname(__file__), '..')
WELCOME = osp.join(PACKAGE_PATH, 'utils', 'templates', 'welcome.html')
WELCOME_DARK = osp.join(PACKAGE_PATH, 'utils', 'templates',
                        'welcome-dark.html')

# Filter to use in file dialogs
FILES_FILTER = '{} (*.ipynb)'.format(_('Jupyter notebooks'))

# How long to wait after save before checking whether file exists (in ms)
WAIT_SAVE_DELAY = 250

# How often to wait for that time
WAIT_SAVE_ITERATIONS = 20

logger = logging.getLogger(__name__)


class NotebookTabWidget(Tabs, SpyderConfigurationAccessor):
    """
    Tabbed widget whose tabs display notebooks.

    This is the main widget of the notebook plugin.

    Attributes
    ----------
    actions : list of (QAction or QMenu or None) or None
        Items to be added to the context menu.
    dark_theme : bool
        Whether to display notebooks in a dark theme. The default is False.
    untitled_num : int
        Number used in file name of newly created notebooks.
    """

    def __init__(self, parent, server_manager, actions=None, menu=None,
                 corner_widgets=None, dark_theme=False):
        """
        Construct a NotebookTabWidget.

        Parameters
        ----------
        parent : QWidget
            Parent of the tabbed widget.
        actions : list of (QAction or QMenu or None) or None
            Items to be added to the context menu.
        menu : QMenu or None
            Context menu of the tabbed widget.
        corner_widgets : dict of (Qt.Corner, list of QWidget or int) or None
            Widgets to be placed in the top left and right corner of the
            tabbed widget. A button for browsing the tabs is always added to
            the top left corner.
        """
        super().__init__(parent, actions, menu, corner_widgets)

        self.actions = actions
        self.dark_theme = dark_theme
        self.untitled_num = 0

        self.server_manager = server_manager
        self.server_manager.sig_server_started.connect(
            self.handle_server_started)
        self.server_manager.sig_server_timed_out.connect(
            self.handle_server_timed_out_or_error)
        self.server_manager.sig_server_errored.connect(
            self.handle_server_timed_out_or_error)

        if not sys.platform == 'darwin':
            # Don't set document mode to true on OSX because it generates
            # a crash when the console is detached from the main window
            # Fixes spyder-ide/spyder#561
            self.setDocumentMode(True)

        self.set_close_function(self.close_client)

    def open_notebook(self, filenames=None):
        """
        Open a notebook from file.

        Parameters
        ----------
        filenames : list of str or None, optional
            List of file names of notebooks to open. The default is None,
            meaning that the user should be asked.

        Returns
        -------
        filenames : list of str
            List of file names of notebooks that were opened.
        """
        if not filenames:
            filenames, _selfilter = getopenfilenames(
                self, _('Open notebook'), '', FILES_FILTER)
        if filenames:
            for filename in filenames:
                self.create_new_client(filename=filename)
        return filenames

    def create_new_client(self, filename=None):
        """
        Create a new notebook or load a pre-existing one.

        Parameters
        ----------
        filename : str, optional
            File name of the notebook to load in the new client. The default
            is None, meaning that a new notebook should be created.

        Returns
        -------
        client : NotebookClient or None
            Notebook client that is opened, or None if unsuccessful.
        """
        # Generate the notebook name (in case of a new one)
        if not filename:
            if not osp.isdir(NOTEBOOK_TMPDIR):
                os.makedirs(NOTEBOOK_TMPDIR)
            nb_name = 'untitled' + str(self.untitled_num) + '.ipynb'
            filename = osp.join(NOTEBOOK_TMPDIR, nb_name)
            kernelspec = dict(display_name='Python 3 (Spyder)',
                              name='python3')
            metadata = dict(kernelspec=kernelspec)
            nb_contents = nbformat.v4.new_notebook(metadata=metadata)
            nbformat.write(nb_contents, filename)
            self.untitled_num += 1

        client = NotebookClient(self, filename, self.actions)
        self.add_tab(client)
        interpreter = self.get_interpreter()
        server_info = self.server_manager.get_server(
            filename, interpreter, start=True)
        if server_info:
            logger.debug('Using existing server at %s',
                         server_info['notebook_dir'])
            client.register(server_info)
            client.load_notebook()
        return client

    def get_interpreter(self):
        """
        Return the Python interpreter to be used in notebooks.

        This function looks in the Spyder configuration to determine and
        return the Python interpreter to be used in notebooks, which is the
        same as is used in consoles.

        Returns
        -------
        The file name of the interpreter
        """
        pyexec = self.get_conf(
            option='executable',
            section='main_interpreter',
            default=get_python_executable()
        )
        return pyexec

    def maybe_create_welcome_client(self):
        """
        Create a welcome tab if there are no tabs.

        Returns
        -------
        client : NotebookClient or None
            The client in the created tab, or None if no tab is created.
        """
        if self.count() == 0:
            if self.dark_theme:
                welcome = open(WELCOME_DARK).read()
            else:
                welcome = open(WELCOME).read()
            client = NotebookClient(
                self, WELCOME, self.actions, ini_message=welcome)
            self.add_tab(client)
            return client

    def close_client(self, index=None, save_before_close=True):
        """
        Close client tab with given index (or close current tab).

        First save the notebook (unless this is the welcome client or
        `save_before_close` is False). Then delete the notebook if it is in
        `get_temp_dir()`. Then shutdown the kernel of the notebook and close
        the tab. Finally, create a welcome tab if there are no tabs.

        Parameters
        ----------
        index : int or None, optional
            Index of tab to be closed. The default is None, meaning that the
            current tab is closed.
        save_before_close : bool, optional
            Whether to save the notebook before closing the tab. The default
            is True.

        Returns
        -------
        The file name of the notebook, or None if no tab was closed.
        """
        if not self.count():
            return None
        if index is None:
            index = self.currentIndex()
        client = self.widget(index)

        filename = client.filename
        if not self.is_welcome_client(client):
            if save_before_close:
                filename = self.save_notebook(client)
            client.shutdown_kernel()
        client.close()

        # Delete notebook file if it is in temporary directory
        if filename.startswith(get_temp_dir()):
            try:
                os.remove(filename)
            except EnvironmentError:
                pass

        # Note: notebook index may have changed after closing related widgets
        self.removeTab(self.indexOf(client))
        self.maybe_create_welcome_client()
        return filename

    def save_notebook(self, client):
        """
        Save notebook corresponding to given client.

        If the notebook is newly created and not empty, then ask the user
        whether to save it under a new name.

        Parameters
        ----------
        client : NotebookClient
            Client of notebook to be saved.

        Returns
        -------
        The file name of the notebook.
        """
        client.save()
        filename = client.filename
        if not self.is_newly_created(client):
            return filename
        if self.wait_and_check_if_empty(filename):
            return filename

        # Notebook not empty, so ask user to save with new filename
        buttons = QMessageBox.Yes | QMessageBox.No
        text = _("<b>{0}</b> has been modified.<br>"
                 "Do you want to save changes?").format(osp.basename(filename))
        answer = QMessageBox.question(
            self, _('Save changes'), text, buttons)
        if answer == QMessageBox.Yes:
            return self.save_as(reopen_after_save=False)
        else:
            return filename

    @staticmethod
    def wait_and_check_if_empty(filename):
        """
        Wait until notebook is created and check whether it is empty.

        Repeatedly try to read the file, waiting a bit after every attempt.
        At the first attempt where the file exists, test whether it is empty
        and return. If it takes too long before the file is created, pretend
        it is empty.

        Parameters
        ----------
        filename : str
            File name of notebook to be checked.

        Returns
        -------
        True if notebook is empty or on timeout, False otherwise.
        """
        for iteration in range(WAIT_SAVE_ITERATIONS):

            # Wait a bit
            wait_save = QEventLoop()
            QTimer.singleShot(WAIT_SAVE_DELAY, wait_save.quit)
            wait_save.exec_()

            # Try reading the file
            try:
                nb_contents = nbformat.read(filename, as_version=4)
            except FileNotFoundError:
                continue

            # If empty, we are done
            if (len(nb_contents['cells']) == 0
                    or len(nb_contents['cells'][0]['source']) == 0):
                return True
            else:
                return False
        else:
            # It is taking longer than expected;
            # Just return True and hope for the best
            return True

    def save_as(self, name=None, reopen_after_save=True):
        """
        Save current notebook under a different file name.

        First, save the notebook under the original file name. Then ask user
        for a new file name (if `name` is not set), and return if no new name
        is given. Then, read the contents of the notebook that was just saved
        and write them under the new file name. If `reopen_after_save` is
        True, then close the original tab and open a new tab with the
        notebook loaded from the new file name.

        Parameters
        ----------
        name : str or None, optional
            File name under which the notebook is to be saved. The default is
            None, meaning that the user should be asked for the file name.
        reopen_after_save : bool, optional
            Whether to close the original tab and re-open it under the new
            file name after saving the notebook. The default is True.

        Returns
        -------
        The file name of the notebook.
        """
        current_client = self.currentWidget()
        current_client.save()
        original_path = current_client.get_filename()
        if not name:
            original_name = osp.basename(original_path)
        else:
            original_name = name
        filename, _selfilter = getsavefilename(self, _("Save notebook"),
                                               original_name, FILES_FILTER)
        if not filename:
            return original_path

        try:
            nb_contents = nbformat.read(original_path, as_version=4)
        except EnvironmentError as error:
            txt = (_("Error while reading {}<p>{}")
                   .format(original_path, str(error)))
            QMessageBox.critical(self, _("File Error"), txt)
            return original_path
        try:
            nbformat.write(nb_contents, filename)
        except EnvironmentError as error:
            txt = (_("Error while writing {}<p>{}")
                   .format(filename, str(error)))
            QMessageBox.critical(self, _("File Error"), txt)
            return original_path
        if reopen_after_save:
            self.close_client(save_before_close=False)
            self.create_new_client(filename=filename)
        return filename

    @staticmethod
    def is_newly_created(client):
        """
        Return whether client has a newly created notebook.

        This only looks at the file name of the notebook. If it has the form
        of file names of newly created notebooks, the function returns True.

        Parameters
        ----------
        client : NotebookClient
            Client under consideration.

        Returns
        -------
        True if notebook is newly created, False otherwise.
        """
        path = client.get_filename()
        dirname, basename = osp.split(path)
        return dirname == NOTEBOOK_TMPDIR and basename.startswith('untitled')

    @staticmethod
    def is_welcome_client(client):
        """
        Return whether some client is a newly created notebook.

        Parameters
        ----------
        client : NotebookClient
            Client under consideration.

        Returns
        -------
        True if `client` is a welcome client, False otherwise.
        """
        return client.get_filename() in [WELCOME, WELCOME_DARK]

    def add_tab(self, widget):
        """
        Add tab containing some notebook widget to the tabbed widget.

        Parameters
        ----------
        widget : NotebookClient
            Notebook widget to display in new tab.
        """
        index = self.addTab(widget, widget.get_short_name())
        self.setCurrentIndex(index)
        self.setTabToolTip(index, widget.get_filename())

    def handle_server_started(self, process):
        """
        Handle signal that a notebook server has started.

        Go through all notebook tabs which do not have server info and try
        getting the server info for them. If that marches the server process
        that has started, then update the notebook tab.

        Parameters
        ----------
        process : ServerProcess
            Info about the server that has started.
        """
        pid = process.process.processId()
        for client_index in range(self.count()):
            client = self.widget(client_index)
            if not client.static and not client.server_url:
                logger.debug('Getting server for %s', client.filename)
                server_info = self.server_manager.get_server(
                    client.filename, process.interpreter, start=False)
                if server_info and server_info['pid'] == pid:
                    logger.debug('Success')
                    client.register(server_info)
                    client.load_notebook()

    def handle_server_timed_out_or_error(self, process):
        """
        Display message box that server failed to start.

        Parameters
        ----------
        process : ServerProcess
            Info about the server that failed to start.
        """
        QMessageBox.critical(
            self,
            _("Server error"),
            _("The Spyder Notebook server failed to start or it is "
              "taking too much time to do it. Please select 'Server info' "
              "in the plugin's option menu to check for errors."))
        # Create a welcome widget, see gh:spyder-ide/spyder-notebook#93
        self.untitled_num -= 1
        self.maybe_create_welcome_client()
        return None



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""
Simple application for working with notebooks.

This is a stand-alone application showing how spyder_notebook can be used
outside Spyder. It is mainly meant for development and testing purposes,
but it can also serve as an example.
"""

# Standard library imports
import argparse
import logging
import sys

# Qt imports
import qdarkstyle
from qtpy.QtCore import QCoreApplication, Qt
from qtpy.QtQuick import QQuickWindow, QSGRendererInterface
from qtpy.QtWidgets import QAction, QApplication, QMainWindow

# Local imports
from spyder_notebook.utils.servermanager import ServerManager
from spyder_notebook.widgets.notebooktabwidget import NotebookTabWidget
from spyder_notebook.widgets.serverinfo import ServerInfoDialog


def use_software_rendering():
    """
    Instruct Qt to use software rendering.

    This is necessary for some buggy graphics drivers (e.g. nvidia).
    This function should be run before the QApplication is created.
    """
    QCoreApplication.setAttribute(Qt.AA_UseSoftwareOpenGL)
    QQuickWindow.setSceneGraphBackend(QSGRendererInterface.Software)


class NotebookAppMainWindow(QMainWindow):
    """Main window for stand-alone notebook application."""

    def __init__(self, options):
        super().__init__()

        if options.dark:
            self.setStyleSheet(qdarkstyle.load_stylesheet_from_environment())

        self.server_manager = ServerManager(options.dark)
        QApplication.instance().aboutToQuit.connect(
            self.server_manager.shutdown_all_servers)

        self.tabwidget = NotebookTabWidget(
            self, self.server_manager, dark_theme=options.dark)

        if options.notebook:
            self.tabwidget.open_notebook(options.notebook)
        else:
            self.tabwidget.maybe_create_welcome_client()

        self.setCentralWidget(self.tabwidget)
        self._setup_menu()

    def view_servers(self):
        """Display server info."""
        dialog = ServerInfoDialog(self.server_manager.servers, parent=self)
        dialog.show()

    def _setup_menu(self):
        file_menu = self.menuBar().addMenu('File')

        new_action = QAction('New Notebook', self)
        new_action.triggered.connect(self.tabwidget.create_new_client)
        file_menu.addAction(new_action)

        open_action = QAction('Open Notebook...', self)
        open_action.triggered.connect(self.tabwidget.open_notebook)
        file_menu.addAction(open_action)

        save_action = QAction('Save Notebook', self)
        save_action.triggered.connect(
            lambda checked: self.tabwidget.save_notebook(
                self.tabwidget.currentWidget()))
        file_menu.addAction(save_action)

        saveas_action = QAction('Save As...', self)
        saveas_action.triggered.connect(self.tabwidget.save_as)
        file_menu.addAction(saveas_action)

        close_action = QAction('Close Notebook', self)
        close_action.triggered.connect(
            lambda checked: self.tabwidget.close_client(
                self.tabwidget.currentIndex()))
        file_menu.addAction(close_action)

        misc_menu = self.menuBar().addMenu('Misc')

        servers_action = QAction('View Servers...', self)
        servers_action.triggered.connect(self.view_servers)
        misc_menu.addAction(servers_action)


def main():
    """Execute example application."""
    use_software_rendering()
    logging.basicConfig(level=logging.DEBUG)
    app = QApplication(sys.argv)
    arguments = app.arguments()
    parser = argparse.ArgumentParser()
    parser.add_argument('notebook', nargs='*')
    parser.add_argument('--dark', action='store_true')
    result = parser.parse_args(arguments[1:])  # arguments[0] = program name
    window = NotebookAppMainWindow(result)
    window.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for the plugin."""

# Standard library imports
import collections
import json
import os
import os.path as osp
import shutil
import sys
from unittest.mock import Mock

# Third-party library imports
from flaky import flaky
import pytest
import requests
from qtpy.QtCore import Qt, QTimer
from qtpy.QtWebEngineWidgets import WEBENGINE
from qtpy.QtWidgets import QFileDialog, QApplication, QLineEdit, QMainWindow

# Spyder imports
from spyder.api.plugins import Plugins
from spyder.config.manager import CONF

# Local imports
from spyder_notebook.notebookplugin import NotebookPlugin

# =============================================================================
# Constants
# =============================================================================
NOTEBOOK_UP = 40000
CALLBACK_TIMEOUT = 10000
INTERACTION_CLICK = 100
LOCATION = osp.realpath(osp.join(os.getcwd(), osp.dirname(__file__)))


# =============================================================================
# Utility functions
# =============================================================================
def prompt_present(nbwidget, qtbot):
    """Check if an In prompt is present in the notebook."""
    return text_present(nbwidget, qtbot, '[ ]:')


def text_present(nbwidget, qtbot, text="Test"):
    """Check if a text is present in the notebook."""
    if WEBENGINE:
        with qtbot.waitCallback(timeout=CALLBACK_TIMEOUT) as cb:
            nbwidget.dom.toHtml(cb)
        return text in cb.args[0]
    else:
        return text in nbwidget.dom.toHtml()


def manage_save_dialog(qtbot, fname, directory=LOCATION):
    """
    Manage the QFileDialog when saving.

    You can use this with QTimer to manage the QFileDialog.
    Before calling anything that may show a QFileDialog for save call:
    QTimer.singleShot(1000, lambda: manage_save_dialog(qtbot))
    """
    top_level_widgets = QApplication.topLevelWidgets()
    for w in top_level_widgets:
        if isinstance(w, QFileDialog):
            if directory is not None:
                w.setDirectory(directory)
            input_field = w.findChildren(QLineEdit)[0]
            input_field.setText(fname)
            qtbot.keyClick(w, Qt.Key_Enter)


def is_kernel_up(kernel_id, sessions_url):
    """Determine if the kernel with the id is up."""
    sessions_req = requests.get(sessions_url).content.decode()
    sessions = json.loads(sessions_req)

    kernel = False
    for session in sessions:
        if kernel_id == session['kernel']['id']:
            kernel = True
            break

    return kernel


# =============================================================================
# Fixtures
# =============================================================================
class MainMock(QMainWindow):
    def __init__(self):
        super().__init__()
        self.switcher = Mock()
        self.main = self
        self.resize(640, 480)

    def get_plugin(self, plugin_name, error=True):
        return Mock()


@pytest.fixture
def notebook(qtbot):
    """Set up the Notebook plugin with a welcome tab and a tab with a new
    notebook. The latter tab is selected."""
    window = MainMock()
    notebook_plugin = NotebookPlugin(parent=window, configuration=CONF)
    notebook_plugin.get_widget().tabwidget.maybe_create_welcome_client()
    notebook_plugin.get_widget().create_new_client()
    window.setCentralWidget(notebook_plugin.get_widget())
    window.show()

    qtbot.addWidget(window)
    yield notebook_plugin
    notebook_plugin.get_widget().close()


@pytest.fixture
def plugin_no_server(mocker, qtbot):
    """Set up the Notebook plugin with a fake nbopen which does not start
    a notebook server."""
    def fake_get_server(filename, interpreter, start):
        return collections.defaultdict(
            str, filename=filename, notebook_dir=osp.dirname(filename))
    fake_server_manager = mocker.Mock(get_server=fake_get_server)
    mocker.patch('spyder_notebook.widgets.main_widget.ServerManager',
                 return_value=fake_server_manager)

    window = MainMock()
    plugin = NotebookPlugin(parent=window, configuration=CONF)
    window.show()

    qtbot.addWidget(window)
    yield plugin
    plugin.get_widget().close()


# =============================================================================
# Tests
# =============================================================================
@flaky(max_runs=3)
def test_shutdown_notebook_kernel(notebook, qtbot):
    """Test that kernel is shutdown from server when closing a notebook."""
    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Get kernel id for the client
    client = notebook.get_widget().tabwidget.currentWidget()
    qtbot.waitUntil(lambda: client.get_kernel_id() is not None,
                    timeout=NOTEBOOK_UP)
    kernel_id = client.get_kernel_id()
    sessions_url = client.get_session_url()

    # Close the current client
    notebook.get_widget().tabwidget.close_client()

    # Assert that the kernel is down for the closed client
    assert not is_kernel_up(kernel_id, sessions_url)


def test_file_in_temp_dir_deleted_after_notebook_closed(notebook, qtbot):
    """Test that notebook file in temporary directory is deleted after the
    notebook is closed."""
    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Get file name
    client = notebook.get_widget().tabwidget.currentWidget()
    filename = client.get_filename()

    # Close the current client
    notebook.get_widget().tabwidget.close_client()

    # Assert file is deleted
    assert not osp.exists(filename)


def test_close_nonexisting_notebook(notebook, qtbot):
    """Test that we can close a tab if the notebook file does not exist.
    Regression test for spyder-ide/spyder-notebook#187."""
    # Set up tab with non-existingg notebook
    filename = osp.join(LOCATION, 'does-not-exist.ipynb')
    notebook.open_notebook(filenames=[filename])
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)
    client = notebook.get_widget().tabwidget.currentWidget()

    # Close tab
    notebook.get_widget().tabwidget.close_client()

    # Assert tab is closed (without raising an exception)
    for client_index in range(notebook.get_widget().tabwidget.count()):
        assert client != notebook.get_widget().tabwidget.widget(client_index)


# TODO Find out what goes wrong on Mac
@flaky(max_runs=3)
@pytest.mark.skipif(sys.platform == 'darwin', reason='Prompt never comes up')
def test_open_notebook_in_non_ascii_dir(notebook, qtbot, tmpdir):
    """Test that a notebook can be opened from a non-ascii directory."""
    # Move the test file to non-ascii directory
    test_notebook = osp.join(LOCATION, 'test.ipynb')
    test_notebook_non_ascii = osp.join(str(tmpdir), u'äöüß', 'test.ipynb')
    os.mkdir(os.path.join(str(tmpdir), u'äöüß'))
    shutil.copyfile(test_notebook, test_notebook_non_ascii)

    # Wait for prompt
    notebook.open_notebook(filenames=[test_notebook_non_ascii])
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Assert that the In prompt has "Test" in it
    # and the client has the correct name
    qtbot.waitUntil(lambda: text_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)
    assert text_present(nbwidget, qtbot)
    assert notebook.get_widget().tabwidget.currentWidget().get_short_name() ==\
           "test"


@flaky(max_runs=3)
@pytest.mark.skipif(not sys.platform.startswith('linux'),
                    reason='Test hangs on CI on Windows and MacOS')
def test_save_notebook(notebook, qtbot, tmpdir):
    """Test that a notebook can be saved."""
    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Writes: a = "test"
    qtbot.keyClick(nbwidget, Qt.Key_A, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_Space, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_Equal, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_Space, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_QuoteDbl, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_T, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_E, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_S, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_T, delay=INTERACTION_CLICK)
    qtbot.keyClick(nbwidget, Qt.Key_QuoteDbl, delay=INTERACTION_CLICK)

    # Save the notebook
    name = osp.join(str(tmpdir), 'save.ipynb')
    QTimer.singleShot(1000, lambda: manage_save_dialog(qtbot, fname=name))
    notebook.get_widget().save_as()

    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Assert that the In prompt has "test" in it
    # and the client has the correct name
    qtbot.waitUntil(lambda: text_present(nbwidget, qtbot, text="test"),
                    timeout=NOTEBOOK_UP)
    assert text_present(nbwidget, qtbot, text="test")
    assert notebook.get_widget().tabwidget.currentWidget().get_short_name() ==\
           "save"


def test_save_notebook_as_with_error(mocker, notebook, qtbot, tmpdir):
    """Test that errors are handled in save_as()."""
    # Set up mocks
    name = osp.join(str(tmpdir), 'save.ipynb')
    mocker.patch('spyder_notebook.widgets.notebooktabwidget.getsavefilename',
                 return_value=(name, 'ignored'))
    mocker.patch('spyder_notebook.widgets.notebooktabwidget.nbformat.write',
                 side_effect=PermissionError)
    mock_critical = mocker.patch('spyder_notebook.widgets.notebooktabwidget'
                                 '.QMessageBox.critical')

    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Save the notebook
    notebook.get_widget().save_as()

    # Assert that message box is displayed (reporting error raised by write)
    assert mock_critical.called


@flaky(max_runs=3)
def test_new_notebook(notebook, qtbot):
    """Test that a new client is really a notebook."""
    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Assert that we have one notebook and the welcome page
    assert notebook.get_widget().tabwidget.count() == 2


# Teardown sometimes fails on Mac with Python 3.8 due to NoProcessException
# in shutdown_server() in notebookapp.py in external notebook library
@flaky
def test_open_console_when_no_kernel(notebook, qtbot, mocker):
    """Test that open_console() handles the case when there is no kernel."""
    # Create mock IPython console plugin and QMessageBox
    MockMessageBox = mocker.patch(
        'spyder_notebook.widgets.main_widget.QMessageBox')

    # Wait for prompt
    nbwidget = notebook.get_widget().tabwidget.currentWidget().notebookwidget
    qtbot.waitUntil(lambda: prompt_present(nbwidget, qtbot),
                    timeout=NOTEBOOK_UP)

    # Shut the kernel down and check that this is successful
    client = notebook.get_widget().tabwidget.currentWidget()
    kernel_id = client.get_kernel_id()
    sessions_url = client.get_session_url()
    client.shutdown_kernel()
    assert not is_kernel_up(kernel_id, sessions_url)

    # Try opening a console
    notebook.get_widget().open_console(client)

    # Assert that a dialog is displayed and no console was opened
    MockMessageBox.critical.assert_called()
    ipyconsole = notebook.get_plugin(Plugins.IPythonConsole)
    ipyconsole._create_client_for_kernel.assert_not_called()


def test_on_mainwindow_visible_with_opened_notebooks(plugin_no_server):
    """
    Run .on_mainwindow_visible() with the `opened_notebooks` conf option set to
    a non-empty list. Check that plugin opens those notebooks.
    """
    plugin = plugin_no_server
    plugin.set_conf('opened_notebooks', ['ham.ipynb', 'spam.ipynb'])

    plugin.on_mainwindow_visible()

    tabwidget = plugin.get_widget().tabwidget
    assert tabwidget.count() == 2
    assert tabwidget.widget(0).filename == 'ham.ipynb'
    assert tabwidget.widget(1).filename == 'spam.ipynb'


def test_on_mainwindow_visible_with_opened_notebooks_empty(plugin_no_server):
    """
    Run .on_mainwindow_visible() with the `opened_notebooks` conf option set to
    an empty list. Check that plugin opens a welcome tab and a new notebook,
    and that the welcome tab is on top.
    """
    plugin = plugin_no_server
    plugin.set_conf('opened_notebooks', [])

    plugin.on_mainwindow_visible()

    tabwidget = plugin.get_widget().tabwidget
    assert tabwidget.count() == 2
    assert tabwidget.is_welcome_client(tabwidget.widget(0))
    assert tabwidget.is_newly_created(tabwidget.widget(1))
    assert tabwidget.currentIndex() == 0


def test_closing_main_widget(mocker, plugin_no_server):
    """Close the main widget with a welcome tab, a new notebooks and a notebook
    opened from a file. Check that config variables `recent_notebooks` and
    `opened_notebook` are correctly set."""
    plugin = plugin_no_server
    main_widget = plugin.get_widget()
    main_widget.clear_recent_notebooks()
    mock_set_option = mocker.patch.object(main_widget, 'set_conf')
    main_widget.tabwidget.maybe_create_welcome_client()
    main_widget.create_new_client()
    main_widget.open_notebook(['ham.ipynb'])

    plugin.get_widget().close()

    expected = [mocker.call('recent_notebooks', ['ham.ipynb']),
                mocker.call('opened_notebooks', ['ham.ipynb'])]
    assert mock_set_option.call_args_list == expected


def test_view_server_info(mocker, plugin_no_server):
    """Check that the "server info" action shows a dialog window with the
    server data."""
    plugin = plugin_no_server
    mock_ServerInfoDialog = mocker.patch(
        'spyder_notebook.widgets.main_widget.ServerInfoDialog')

    plugin.get_widget().server_info_action.trigger()

    mock_ServerInfoDialog.assert_called_once_with(
        plugin.get_widget().server_manager.servers, parent=plugin.get_widget())
    mock_ServerInfoDialog.return_value.show.assert_called_once()


@pytest.mark.parametrize('config_value', ['same as spyder', 'dark', 'light'])
@pytest.mark.parametrize('spyder_is_dark', [True, False])
def test_dark_theme(mocker, plugin_no_server, config_value, spyder_is_dark):
    plugin_no_server.set_conf('theme', config_value)
    mocker.patch('spyder_notebook.widgets.main_widget.is_dark_interface',
                 return_value=spyder_is_dark)

    value = plugin_no_server.get_widget().dark_theme

    expected = (config_value == 'dark' or
                (config_value == 'same as spyder' and spyder_is_dark))
    assert value == expected


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Tests for notebooktabwidget.py."""

# Standard library imports
import collections
import os.path as osp

# Third party imports
import pytest
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder_notebook.utils.servermanager import ServerManager
from spyder_notebook.widgets.notebooktabwidget import (
    NotebookTabWidget, WAIT_SAVE_ITERATIONS)


@pytest.fixture
def tabwidget(mocker, qtbot):
    """Create an empty NotebookTabWidget which does not start up servers."""
    def fake_get_server(filename, interpreter, start):
        return collections.defaultdict(
            str, filename=filename, notebook_dir=osp.dirname(filename))

    mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.WAIT_SAVE_DELAY', 1)
    fake_server_manager = mocker.Mock(
        spec=ServerManager, get_server=fake_get_server)
    widget = NotebookTabWidget(None, fake_server_manager)
    qtbot.addWidget(widget)
    return widget


def test_save_notebook_with_opened_notebook(mocker, tabwidget):
    """Test that .save_notebook() with a notebook opened from a file does
    indeed save the notebook but does not check whether it is empty."""
    client = tabwidget.create_new_client('ham.ipynb')
    client.save = mocker.Mock()
    tabwidget.wait_and_check_if_empty = mocker.Mock()

    result = tabwidget.save_notebook(client)

    client.save.assert_called()
    tabwidget.wait_and_check_if_empty.assert_not_called()
    assert result == 'ham.ipynb'


def test_save_notebook_with_empty_new_notebook(mocker, tabwidget):
    """Test that .save_notebook() on a newly created notebook saves the
    notebook, checks that it is empty, and it is, does not ask to save it
    again."""
    mock_question = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.QMessageBox.question')
    client = tabwidget.create_new_client()
    client.save = mocker.Mock()
    tabwidget.wait_and_check_if_empty = mocker.Mock(return_value=True)

    result = tabwidget.save_notebook(client)

    client.save.assert_called()
    tabwidget.wait_and_check_if_empty.assert_called()
    mock_question.assert_not_called()
    assert result.endswith('untitled0.ipynb')


def test_save_notebook_with_nonempty_new_notebook_and_save(mocker, tabwidget):
    """Test that .save_notebook() on a newly created notebook saves the
    notebook, checks that it is empty, and if it is, asks to save it again,
    and if yes, does save it under a new name."""
    mock_question = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.QMessageBox.question',
        return_value=QMessageBox.Yes)
    client = tabwidget.create_new_client()
    client.save = mocker.Mock()
    tabwidget.wait_and_check_if_empty = mocker.Mock(return_value=False)
    tabwidget.save_as = mocker.Mock(return_value='newname.ipynb')

    result = tabwidget.save_notebook(client)

    client.save.assert_called()
    tabwidget.wait_and_check_if_empty.assert_called_once()
    mock_question.assert_called_once()
    tabwidget.save_as.assert_called_once()
    assert result == 'newname.ipynb'


def test_save_notebook_with_nonempty_new_notebook_and_no_save(
            mocker, tabwidget):
    """Test that .save_notebook() on a newly created notebook saves the
    notebook, checks that it is empty, and if it is, asks to save it again,
    and if no, does not save it under a new name."""
    mock_question = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.QMessageBox.question',
        return_value=QMessageBox.No)
    client = tabwidget.create_new_client()
    client.save = mocker.Mock()
    tabwidget.wait_and_check_if_empty = mocker.Mock(return_value=False)
    tabwidget.save_as = mocker.Mock(return_value='newname.ipynb')

    result = tabwidget.save_notebook(client)

    client.save.assert_called()
    tabwidget.wait_and_check_if_empty.assert_called_once()
    mock_question.assert_called_once()
    tabwidget.save_as.assert_not_called()
    assert result.endswith('untitled0.ipynb')


def test_wait_and_check_if_empty_when_empty(mocker, tabwidget):
    """Test that .wait_and_check_if_empty() returns True when called on a
    notebook that is empty."""
    contents = {'cells': []}
    mock_read = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.nbformat.read',
        return_value=contents)

    result = tabwidget.wait_and_check_if_empty('ham.ipynb')

    mock_read.assert_called_once()
    assert result is True


def test_wait_and_check_if_empty_when_not_empty(mocker, tabwidget):
    """Test that .wait_and_check_if_empty() returns False when called on a
    notebook that is not empty."""
    contents = {'cells': [{'source': 'not empty'}]}
    mock_read = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.nbformat.read',
        return_value=contents)

    result = tabwidget.wait_and_check_if_empty('ham.ipynb')

    mock_read.assert_called_once()
    assert result is False


def test_wait_and_check_if_empty_with_delay(mocker, tabwidget):
    """Test that .wait_and_check_if_empty() on an empty notebook tries to read
    it, and when that fails because the file does not exist, it tries again to
    read it. When the read succeeds the second time and the notebook turns out
    to be empty, the function returns True."""
    contents = {'cells': []}
    mock_read = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.nbformat.read',
        side_effect=[FileNotFoundError, contents])

    result = tabwidget.wait_and_check_if_empty('ham.ipynb')

    assert mock_read.call_count == 2
    assert result is True


def test_wait_and_check_with_long_delay(mocker, tabwidget):
    """Test that .wait_and_check_if_empty() repeatedly tries to read the file,
    and when that keeps failing because the file does not exists, eventually
    gives up and returns True."""
    mock_read = mocker.patch(
        'spyder_notebook.widgets.notebooktabwidget.nbformat.read',
        side_effect=FileNotFoundError)

    result = tabwidget.wait_and_check_if_empty('ham.ipynb')

    assert mock_read.call_count == WAIT_SAVE_ITERATIONS
    assert result is True


def test_is_newly_created_with_new_notebook(tabwidget):
    """Test that .is_newly_created() returns True if passed a client that is
    indeed newly created."""
    client = tabwidget.create_new_client()
    assert tabwidget.is_newly_created(client)


def test_is_newly_created_with_opened_notebook(tabwidget):
    """Test that .is_newly_created() returns False if passed a client that
    contains a notebook opened from a file."""
    client = tabwidget.create_new_client('ham.ipynb')
    assert not tabwidget.is_newly_created(client)


def test_is_newly_created_with_welcome_tab(tabwidget):
    """Test that .is_newly_created() returns False if passed a welcome
    client."""
    client = tabwidget.maybe_create_welcome_client()
    assert not tabwidget.is_newly_created(client)


def test_is_welcome_client_with_new_notebook(tabwidget):
    """Test that .is_welcome_client() returns False if passed a client that is
    indeed newly created."""
    client = tabwidget.create_new_client()
    assert not tabwidget.is_welcome_client(client)


def test_is_welcome_client_with_opened_notebook(tabwidget):
    """Test that .is_welcome_client() returns False if passed a client that
    contains a notebook opened from a file."""
    client = tabwidget.create_new_client('ham.ipynb')
    assert not tabwidget.is_welcome_client(client)


def test_is_welcome_client_with_welcome_tab(tabwidget):
    """Test that .is_welcome_client() returns True if passed a welcome
    client."""
    client = tabwidget.maybe_create_welcome_client()
    assert tabwidget.is_welcome_client(client)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Tests for client.py covering NotebookClient."""

# Third-party imports
import pytest
from qtpy.QtCore import QUrl
from qtpy.QtWidgets import QWidget
import requests

# Local imports
from spyder_notebook.widgets.client import NotebookClient


class MockPlugin(QWidget):
    def get_plugin_actions(self):
        return []


@pytest.fixture
def plugin_without_server(qtbot):
    """
    Construct mock plugin with NotebookClient but no notebook server.

    Use `plugin.client` to access the client.
    """
    plugin = MockPlugin()
    qtbot.addWidget(plugin)
    client = NotebookClient(plugin, '/path/notebooks/ham.ipynb')
    plugin.client = client
    return plugin


@pytest.fixture
def plugin(plugin_without_server):
    """
    Construct mock plugin with NotebookClient with server registered.

    Use `plugin.client` to access the client.
    """
    server_info = {'notebook_dir': '/path/notebooks',
                   'url': 'fake_url',
                   'token': 'fake_token'}
    plugin_without_server.client.register(server_info)
    return plugin_without_server


def test_notebookwidget_create_window(plugin, mocker, qtbot):
    """Test that NotebookWidget.create_window() creates an object that calls
    webbrowser.open() when its URL is set."""
    widget = plugin.client.notebookwidget
    mock_notebook_open = mocker.patch('webbrowser.open')
    new_view = widget.createWindow(42)
    url = 'https://www.spyder-ide.org/'

    with qtbot.waitSignal(new_view.loadFinished):
        new_view.setUrl(QUrl(url))

    mock_notebook_open.assert_called_once_with(url)


def test_notebookclient_get_kernel_id(plugin, mocker):
    """Basic unit test for NotebookClient.get_kernel_id()."""
    response = mocker.Mock()
    content = b'[{"kernel": {"id": "42"}, "notebook": {"path": "ham.ipynb"}}]'
    response.content = content
    response.status_code = requests.codes.ok
    mocker.patch('requests.get', return_value=response)

    kernel_id = plugin.client.get_kernel_id()
    assert kernel_id == '42'


def test_notebookclient_get_kernel_id_without_server(
        plugin_without_server, mocker):
    """Test NotebookClient.get_kernel_id() if client has no server."""
    mock_get = mocker.patch('requests.get')

    kernel_id = plugin_without_server.client.get_kernel_id()

    assert kernel_id is None
    mock_get.assert_not_called()


def test_notebookclient_get_kernel_id_with_fields_missing(plugin, mocker):
    """Test NotebookClient.get_kernel_id() if response has fields missing."""
    response = mocker.Mock()
    content = (b'[{"kernel": {"id": "1"}, "notebook": {"spam": "eggs"}},'
               b' {"kernel": {"id": "2"}},'
               b' {"kernel": {"id": "3"}, "notebook": {"path": "ham.ipynb"}}]')
    response.content = content
    response.status_code = requests.codes.ok
    mocker.patch('requests.get', return_value=response)

    kernel_id = plugin.client.get_kernel_id()
    assert kernel_id == '3'


def test_notebookclient_get_kernel_id_with_error_status(plugin, mocker):
    """Test NotebookClient.get_kernel_id() when response has error status.
    In this case, the content of the response may be empty;
    see spyder-ide/spyder-notebook#317."""
    response = mocker.Mock()
    content = b''
    response.content = content
    response.status_code = requests.codes.forbidden
    mocker.patch('requests.get', return_value=response)
    MockMessageBox = mocker.patch('spyder_notebook.widgets.client.QMessageBox')

    plugin.client.get_kernel_id()

    MockMessageBox.warning.assert_called()


def test_notebookclient_get_kernel_id_with_exception(plugin, mocker):
    """Test NotebookClient.get_kernel_id() when request raises an exception."""
    exception = requests.exceptions.ProxyError('kaboom')
    mocker.patch('requests.get', side_effect=exception)
    MockMessageBox = mocker.patch('spyder_notebook.widgets.client.QMessageBox')

    plugin.client.get_kernel_id()

    MockMessageBox.warning.assert_called()



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""ServerInfoDialog, a dialog window showing info about notebook servers."""


# Standard library imports
import sys

# Qt imports
from qtpy.QtWidgets import (
    QApplication, QComboBox, QDialogButtonBox, QFormLayout, QLineEdit,
    QPushButton, QTextEdit, QVBoxLayout)

# Spyder imports
from spyder.plugins.variableexplorer.widgets.basedialog import BaseDialog

# Local imports
from spyder_notebook.utils.localization import _
from spyder_notebook.utils.servermanager import ServerProcess, ServerState


# Text description of the different server states
SERVER_STATE_DESCRIPTIONS = {
    ServerState.STARTING:  _('Starting up'),
    ServerState.RUNNING:   _('Running'),
    ServerState.FINISHED:  _('Finished'),
    ServerState.ERROR:     _('Error'),
    ServerState.TIMED_OUT: _('Timed out')}


class ServerInfoDialog(BaseDialog):
    """Dialog window showing information about notebook servers."""

    def __init__(self, server_info, parent=None):
        """
        Construct a RecoveryDialog.

        Parameters
        ----------
        servers : list of ServerProcess
            Information to be displayed. This parameter is read only.
        parent : QWidget, optional
            Parent of the dialog window. The default is None.
        """
        super().__init__(parent)
        self.servers = server_info

        self.setWindowTitle(_('Notebook server info'))

        self.layout = QVBoxLayout(self)
        self.formlayout = QFormLayout()
        self.layout.addLayout(self.formlayout)

        self.process_combo = QComboBox(self)
        self.process_combo.currentIndexChanged.connect(self.select_process)
        self.formlayout.addRow(_('Process ID:'), self.process_combo)

        self.dir_lineedit = QLineEdit(self)
        self.dir_lineedit.setReadOnly(True)
        self.formlayout.addRow(_('Notebook dir:'), self.dir_lineedit)

        self.interpreter_lineedit = QLineEdit(self)
        self.interpreter_lineedit.setReadOnly(True)
        self.formlayout.addRow(_('Python interpreter:'),
                               self.interpreter_lineedit)

        self.state_lineedit = QLineEdit(self)
        self.state_lineedit.setReadOnly(True)
        self.formlayout.addRow(_('State:'), self.state_lineedit)

        self.log_textedit = QTextEdit(self)
        self.log_textedit.setReadOnly(True)
        self.layout.addWidget(self.log_textedit)

        self.buttonbox = QDialogButtonBox(QDialogButtonBox.Ok, self)
        self.buttonbox.accepted.connect(self.accept)
        self.refresh_button = QPushButton(_('Refresh'), self)
        self.refresh_button.clicked.connect(self.refresh_data)
        self.buttonbox.addButton(
            self.refresh_button, QDialogButtonBox.ActionRole)
        self.layout.addWidget(self.buttonbox)

        self.refresh_data()

    def refresh_data(self):
        self.process_combo.clear()
        for server in self.servers:
            self.process_combo.addItem(str(server.process.processId()))
        self.select_process(0)

    def select_process(self, index):
        self.dir_lineedit.setText(self.servers[index].notebook_dir)
        self.interpreter_lineedit.setText(self.servers[index].interpreter)
        self.state_lineedit.setText(
            SERVER_STATE_DESCRIPTIONS[self.servers[index].state])
        self.log_textedit.setPlainText(self.servers[index].output)


def test():  # pragma: no cover
    """Display dialog for manual testing."""
    class FakeProcess:
        def __init__(self, pid):
            self.pid = pid

        def processId(self):
            return self.pid

    servers = [ServerProcess(FakeProcess(42), '/my/home/dir',
                             '/ham/interpreter',
                             state=ServerState.RUNNING,
                             output='Nicely humming along...\n'),
               ServerProcess(FakeProcess(404), '/some/other/dir',
                             '/spam/interpreter',
                             state=ServerState.FINISHED,
                             output='Terminated for some reason...\n')]

    app = QApplication(sys.argv)
    dialog = ServerInfoDialog(servers)
    dialog.show()
    sys.exit(app.exec_())


if __name__ == "__main__":  # pragma: no cover
    test()



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Notebook plugin."""

# Qt imports
from qtpy.QtCore import Signal

# Spyder imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)

# Local imports
from spyder_notebook.config import CONF_DEFAULTS, CONF_VERSION
from spyder_notebook.confpage import NotebookConfigPage
from spyder_notebook.widgets.main_widget import NotebookMainWidget
from spyder_notebook.utils.localization import _


class NotebookPlugin(SpyderDockablePlugin):
    """Spyder Notebook plugin."""

    NAME = 'notebook'
    REQUIRES = [Plugins.Preferences]
    OPTIONAL = [Plugins.IPythonConsole]
    TABIFY = [Plugins.Editor]
    CONF_SECTION = NAME
    CONF_DEFAULTS = CONF_DEFAULTS
    CONF_VERSION = CONF_VERSION
    WIDGET_CLASS = NotebookMainWidget
    CONF_WIDGET_CLASS = NotebookConfigPage

    # ---- Signals
    # ------------------------------------------------------------------------
    focus_changed = Signal()

    # ---- SpyderDockablePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        """Return plugin name."""
        title = _('Notebook')
        return title

    def get_description(self):
        """Return the plugin description."""
        return _("Work with Jupyter notebooks inside Spyder.")

    def get_icon(self):
        """Return plugin icon."""
        return self.create_icon('notebook')

    def on_initialize(self):
        """Register plugin in Spyder's main window."""
        self.focus_changed.connect(self.main.plugin_focus_changed)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.IPythonConsole)
    def on_ipyconsole_available(self):
        self.get_widget().sig_open_console_requested.connect(
            self._open_console)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.IPythonConsole)
    def on_ipyconsole_teardown(self):
        self.get_widget().sig_open_console_requested.disconnect(
            self._open_console)

    def on_mainwindow_visible(self):
        self.get_widget().open_previous_session()

    # ------ Public API -------------------------------------------------------
    def open_notebook(self, filenames=None):
        self.get_widget().open_notebook(filenames)

    # ------ Private API ------------------------------------------------------
    def _open_console(self, kernel_id, tab_name):
        """Open an IPython console as requested."""
        ipyconsole = self.get_plugin(Plugins.IPythonConsole)
        ipyconsole.create_client_for_kernel(kernel_id)
        ipyclient = ipyconsole.get_current_client()
        ipyclient.allow_rename = False
        ipyconsole.rename_client_tab(ipyclient, tab_name)



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Localization utility function."""

from spyder.config.base import get_translation
_ = get_translation('spyder_notebook')



# -*- coding: utf-8 -*-
#
# Copyright (c) Jupyter Development Team, Spyder Project Contributors.
# Distributed under the terms of the Modified BSD License.

"""Notebook server for the Spyder notebook plugin."""



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for config.py"""

# Third-party library imports
import pytest

# Spyder imports
from spyder.plugins.preferences.widgets.configdialog import ConfigDialog

# Local imports
from spyder_notebook.confpage import NotebookConfigPage
from spyder_notebook.tests.test_plugin import plugin_no_server


def test_config(plugin_no_server, qtbot):
    """Test that config page can be created and shown."""
    dlg = ConfigDialog()
    page = NotebookConfigPage(plugin_no_server, parent=plugin_no_server.main)
    page.initialize()
    dlg.add_page(page)
    dlg.show()
    qtbot.addWidget(dlg)
    # no assert, just check that the config page can be created


if __name__ == "__main__":
    pytest.main()



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Spyder configuration page for notebook plugin."""

# Qt imports
from qtpy.QtWidgets import QGridLayout, QGroupBox, QVBoxLayout

# Spyder imports
from spyder.api.preferences import PluginConfigPage

# Local imports
from spyder_notebook.utils.localization import _


class NotebookConfigPage(PluginConfigPage):
    """Widget with configuration options for notebook plugin."""

    def setup_page(self):
        """Fill configuration page with widgets."""
        themes = ['Same as Spyder', 'Light', 'Dark']
        theme_choices = list(zip(themes,
                                 [theme.lower() for theme in themes]))
        theme_combo = self.create_combobox(
            _('Notebook theme'), theme_choices, 'theme', restart=True)

        interface_layout = QGridLayout()
        interface_layout.addWidget(theme_combo.label, 0, 0)
        interface_layout.addWidget(theme_combo.combobox, 0, 1)
        interface_group = QGroupBox(_('Interface'))
        interface_group.setLayout(interface_layout)

        vlayout = QVBoxLayout()
        vlayout.addWidget(interface_group)
        vlayout.addStretch(1)
        self.setLayout(vlayout)



# -*- coding: utf-8 -*-
#
# Copyright (c) Spyder Project Contributors
# Licensed under the terms of the MIT License

"""Qt widgets for the notebook."""

# Standard library imports
import json
import os
import os.path as osp
from string import Template
import sys

# Third-party imports
from notebook.utils import url_path_join, url_escape
import qstylizer
from qtpy.QtCore import QEvent, QUrl, Qt, Signal
from qtpy.QtGui import QColor, QFontMetrics, QFont
from qtpy.QtWebEngineWidgets import (QWebEnginePage, QWebEngineSettings,
                                     QWebEngineView, WEBENGINE)
from qtpy.QtWidgets import (QApplication, QMenu, QFrame, QVBoxLayout,
                            QMessageBox)
import requests

# Spyder imports
from spyder.config.base import get_module_source_path
from spyder.utils import sourcecode
from spyder.utils.image_path_manager import get_image_path
from spyder.utils.qthelpers import add_actions
from spyder.utils.palette import QStylePalette
from spyder.widgets.findreplace import FindReplace

# Local imports
from spyder_notebook.config import CONF_SECTION
from spyder_notebook.utils.localization import _
from spyder_notebook.widgets.dom import DOMWidget

# -----------------------------------------------------------------------------
# Templates
# -----------------------------------------------------------------------------
# Using the same css file from the Help plugin for now. Maybe
# later it'll be a good idea to create a new one.

PLUGINS_PATH = get_module_source_path('spyder', 'plugins')
TEMPLATES_PATH = osp.join(
    PLUGINS_PATH, 'ipythonconsole', 'assets', 'templates')

BLANK = open(osp.join(TEMPLATES_PATH, 'blank.html')).read()
LOADING = open(osp.join(TEMPLATES_PATH, 'loading.html')).read()
KERNEL_ERROR = open(osp.join(TEMPLATES_PATH, 'kernel_error.html')).read()


# -----------------------------------------------------------------------------
# Widgets
# -----------------------------------------------------------------------------
class WebViewInBrowser(QWebEngineView):
    """
    WebView which opens document in an external browser.

    This is a subclass of QWebEngineView, which as soon as the URL is set,
    opens the web page in an external browser and closes itself. It is used
    in NotebookWidget to open links.
    """

    def __init__(self, parent):
        """Construct object."""
        super().__init__(parent)
        self.urlChanged.connect(self.open_in_browser)

    def open_in_browser(self, url):
        """
        Open web page in external browser and close self.

        Parameters
        ----------
        url : QUrl
            URL of web page to open in browser
        """
        import webbrowser
        try:
            webbrowser.open(url.toString())
        except ValueError:
            # See: spyder-ide/spyder#9849
            pass
        self.stop()
        self.close()


class NotebookWidget(DOMWidget):
    """WebView widget for notebooks."""

    sig_focus_in_event = Signal()
    """
    This signal is emitted when the widget receives focus.
    """

    sig_focus_out_event = Signal()
    """
    This signal is emitted when the widget loses focus.
    """

    def __init__(self, parent, actions=None):
        """
        Constructor.

        Parameters
        ----------
        parent : QWidget
            Parent of the widget under construction.
        actions : list of (QAction or QMenu or None) or None, optional
            Actions to be added to the context menu of the widget under
            construction. The default is None, meaning that no actions
            will be added.
        """
        super().__init__(parent)
        self.CONTEXT_NAME = str(id(self))
        self.setup()
        self.actions = actions

        # Path for css files in Spyder according to the interface theme set by
        # the user (i.e. dark or light).
        self.css_path = self.get_conf('css_path', section='appearance')

        # Set default background color.
        self.page().setBackgroundColor(
            QColor(QStylePalette.COLOR_BACKGROUND_1)
        )

    def contextMenuEvent(self, event):
        """
        Handle context menu events.

        This overrides WebView.contextMenuEvent() in order to add the
        actions in `self.actions` and remove the Back and Forward actions
        which have no meaning for the notebook widget.

        If Shift is pressed, then instead display the standard Qt context menu,
        per gh:spyder-ide/spyder-notebook#279

        Parameters
        ----------
        event : QContextMenuEvent
            The context menu event that needs to be handled.
        """
        if QApplication.keyboardModifiers() & Qt.ShiftModifier:
            return QWebEngineView.contextMenuEvent(self, event)

        if self.actions is None:
            actions = []
        else:
            actions = self.actions + [None]

        actions += [
            self.pageAction(QWebEnginePage.SelectAll),
            self.pageAction(QWebEnginePage.Copy),
            None,
            self.zoom_in_action,
            self.zoom_out_action]

        if not WEBENGINE:
            settings = self.page().settings()
            settings.setAttribute(QWebEngineSettings.DeveloperExtrasEnabled,
                                  True)
            actions += [None, self.pageAction(QWebEnginePage.InspectElement)]

        menu = QMenu(self)
        add_actions(menu, actions)
        menu.popup(event.globalPos())
        event.accept()

    def _set_info(self, html):
        """Set informational html with css from local path."""
        self.setHtml(html, QUrl.fromLocalFile(self.css_path))

    def show_blank(self):
        """Show a blank page."""
        blank_template = Template(BLANK)
        page = blank_template.substitute(css_path=self.css_path)
        self._set_info(page)

    def show_kernel_error(self, error):
        """Show kernel initialization errors."""
        # Remove unneeded blank lines at the beginning
        eol = sourcecode.get_eol_chars(error)
        if eol:
            error = error.replace(eol, '<br>')
        # Don't break lines in hyphens
        # From http://stackoverflow.com/q/7691569/438386
        error = error.replace('-', '&#8209')

        message = _("An error occurred while starting the kernel")
        kernel_error_template = Template(KERNEL_ERROR)
        page = kernel_error_template.substitute(css_path=self.css_path,
                                                message=message,
                                                error=error)
        self._set_info(page)

    def show_loading_page(self):
        """Show a loading animation while the kernel is starting."""
        loading_template = Template(LOADING)
        loading_img = get_image_path('loading_sprites.png')
        if os.name == 'nt':
            loading_img = loading_img.replace('\\', '/')
        message = _("Connecting to kernel...")
        page = loading_template.substitute(css_path=self.css_path,
                                           loading_img=loading_img,
                                           message=message)
        self._set_info(page)

    def show_message(self, page):
        """Show a message page with the given .html file."""
        self.setHtml(page)

    def createWindow(self, webWindowType):
        """
        Create new browser window.

        This function is called by Qt if the user clicks on a link in the
        notebook. The goal is to open the web page in an external browser.
        To that end, we create and return an object which will open the browser
        when Qt sets the URL.
        """
        return WebViewInBrowser(self.parent())

    def eventFilter(self, widget, event):
        """
        Handle events that affect the view.
        All events (e.g. focus in/out) reach the focus proxy, not this
        widget itself. That's why this event filter is necessary.
        """
        if self.focusProxy() is widget:
            if event.type() == QEvent.FocusIn:
                self.sig_focus_in_event.emit()
            elif event.type() == QEvent.FocusOut:
                self.sig_focus_out_event.emit()
        return super().eventFilter(widget, event)


class NotebookClient(QFrame):
    """
    Notebook client for Spyder.

    This is a widget composed of a NotebookWidget and a find dialog to
    render notebooks.

    Attributes
    ----------
    server_url : str or None
        URL to send requests to; set by register().
    """

    CONF_SECTION = CONF_SECTION

    def __init__(self, parent, filename, actions=None, ini_message=None):
        """
        Constructor.

        Parameters
        ----------
        parent : QWidget
            Parent of the widget under construction.
        filename : str
            File name of the notebook.
        actions : list of (QAction or QMenu or None) or None, optional
            Actions to be added to the context menu of the widget under
            construction. The default is None, meaning that no actions
            will be added.
        ini_message : str or None, optional
            HTML to be initially displayed in the widget. The default is
            None, meaning that an empty page is displayed initially.
        """
        super().__init__(parent)

        if os.name == 'nt':
            filename = filename.replace('/', '\\')
        self.filename = filename

        self.file_url = None
        self.server_url = None
        self.path = None

        self.notebookwidget = NotebookWidget(self, actions)
        if ini_message:
            self.notebookwidget.show_message(ini_message)
            self.static = True
        else:
            self.notebookwidget.show_blank()
            self.static = False

        self.notebookwidget.sig_focus_in_event.connect(
            lambda: self._apply_stylesheet(focus=True))
        self.notebookwidget.sig_focus_out_event.connect(
            lambda: self._apply_stylesheet(focus=False))
        self._apply_stylesheet()

        self.find_widget = FindReplace(self)
        self.find_widget.set_editor(self.notebookwidget)
        self.find_widget.hide()

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.notebookwidget)
        layout.addWidget(self.find_widget)
        self.setLayout(layout)

    def add_token(self, url):
        """Add notebook token to a given url."""
        token_url = url + '?token={}'.format(self.token)
        return token_url

    def register(self, server_info):
        """Register attributes that can be computed with the server info."""
        # Path relative to the server directory
        self.path = os.path.relpath(self.filename,
                                    start=server_info['notebook_dir'])

        # Replace backslashes on Windows
        if os.name == 'nt':
            self.path = self.path.replace('\\', '/')

        # Server url to send requests to
        self.server_url = server_info['url']

        # Server token
        self.token = server_info['token']

        url = url_path_join(self.server_url, 'notebook',
                            url_escape(self.path))

        # Set file url to load this notebook
        self.file_url = self.add_token(url)

    def go_to(self, url_or_text):
        """Go to page URL."""
        if isinstance(url_or_text, str):
            url = QUrl(url_or_text)
        else:
            url = url_or_text
        self.notebookwidget.load(url)

    def load_notebook(self):
        """Load the associated notebook."""
        self.go_to(self.file_url)

    def get_filename(self):
        """Get notebook's filename."""
        return self.filename

    def get_short_name(self):
        """Get a short name for the notebook."""
        sname = osp.splitext(osp.basename(self.filename))[0]
        if len(sname) > 20:
            fm = QFontMetrics(QFont())
            sname = fm.elidedText(sname, Qt.ElideRight, 110)
        return sname

    def save(self):
        """
        Save current notebook asynchronously.

        This function simulates a click on the Save button in the notebook
        which will save the current notebook (but the function will return
        before). The Save button is found by selecting the first element of
        class `jp-ToolbarButtonComponent` whose `title` attribute begins with
        the string "Save".
        """
        self.notebookwidget.mousedown(
            '.jp-ToolbarButtonComponent[title^="Save"]')

    def get_session_url(self):
        """
        Get the kernel sessions URL of the client.

        Return a str with the URL or None, if no server is associated to
        the client.
        """
        if self.server_url:
            session_url = url_path_join(self.server_url, 'api/sessions')
            return self.add_token(session_url)
        else:
            return None

    def get_kernel_id(self):
        """
        Get the kernel id of the client.

        Return a str with the kernel id or None. On error, display a dialog
        box and return None.
        """
        sessions_url = self.get_session_url()
        if not sessions_url:
            return None

        try:
            sessions_response = requests.get(sessions_url)
        except requests.exceptions.RequestException as exception:
            msg = _('Spyder could not get a list of sessions '
                    'from the Jupyter Notebook server. '
                    'Message: {}').format(exception)
            QMessageBox.warning(self, _('Server error'), msg)
            return None

        if sessions_response.status_code != requests.codes.ok:
            msg = _('Spyder could not get a list of sessions '
                    'from the Jupyter Notebook server. '
                    'Status code: {}').format(sessions_response.status_code)
            QMessageBox.warning(self, _('Server error'), msg)
            return None

        if os.name == 'nt':
            path = self.path.replace('\\', '/')
        else:
            path = self.path

        sessions = json.loads(sessions_response.content.decode())
        for session in sessions:
            notebook_path = session.get('notebook', {}).get('path')
            if notebook_path is not None and notebook_path == path:
                kernel_id = session['kernel']['id']
                return kernel_id

    def shutdown_kernel(self):
        """Shutdown the kernel of the client."""
        kernel_id = self.get_kernel_id()

        if kernel_id:
            delete_url = self.add_token(url_path_join(self.server_url,
                                                      'api/kernels/',
                                                      kernel_id))
            delete_req = requests.delete(delete_url)
            if delete_req.status_code != 204:
                QMessageBox.warning(
                    self,
                    _("Server error"),
                    _("The Jupyter Notebook server "
                      "failed to shutdown the kernel "
                      "associated with this notebook. "
                      "If you want to shut it down, "
                      "you'll have to close Spyder."))

    def _apply_stylesheet(self, focus=False):
        """Apply stylesheet according to the current focus."""
        if focus:
            border_color = QStylePalette.COLOR_ACCENT_3
        else:
            border_color = QStylePalette.COLOR_BACKGROUND_4

        css = qstylizer.style.StyleSheet()
        css.QFrame.setValues(
            border=f'1px solid {border_color}',
            margin='0px 1px 0px 1px',
            padding='0px 0px 1px 0px',
            borderRadius='3px'
        )

        self.setStyleSheet(css.toString())


# -----------------------------------------------------------------------------
# Tests
# -----------------------------------------------------------------------------
def main():
    """Execute a simple test."""
    from spyder.utils.qthelpers import qapplication
    app = qapplication()
    widget = NotebookClient(parent=None, filename='')
    widget.show()
    widget.go_to('http://google.com')
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for serverinfo.py."""

# Third party imports
import pytest

# Local imports
from spyder_notebook.utils.servermanager import ServerProcess, ServerState
from spyder_notebook.widgets.serverinfo import ServerInfoDialog


class FakeProcess:
    """Fake for ServerProcess class."""

    def __init__(self, pid):
        self.pid = pid

    def processId(self):
        """Member function which needs to be faked."""
        return self.pid


@pytest.fixture
def dialog(qtbot):
    """Construct and return dialog window for testing."""
    servers = [ServerProcess(FakeProcess(42), '/my/home/dir',
                             interpreter='/ham/interpreter',
                             state=ServerState.RUNNING,
                             output='Nicely humming along...\n'),
               ServerProcess(FakeProcess(404), '/some/other/dir',
                             interpreter='/spam/interpreter',
                             state=ServerState.FINISHED,
                             output='Terminated for some reason...\n')]
    res = ServerInfoDialog(servers)
    qtbot.addWidget(res)
    return res


def test_dialog_on_initialization(dialog):
    """Test that dialog's data is filled correctly after initialization."""
    assert dialog.process_combo.count() == 2
    assert dialog.process_combo.currentIndex() == 0
    assert dialog.process_combo.currentText() == '42'
    assert dialog.process_combo.itemText(1) == '404'
    assert dialog.state_lineedit.text() == 'Running'
    assert dialog.dir_lineedit.text() == '/my/home/dir'
    assert dialog.interpreter_lineedit.text() == '/ham/interpreter'
    assert dialog.log_textedit.toPlainText() == 'Nicely humming along...\n'


def test_dialog_change_process(dialog):
    """Test that dialog's data changes when user select another process."""
    dialog.process_combo.setCurrentIndex(1)

    assert dialog.process_combo.count() == 2
    assert dialog.process_combo.currentIndex() == 1
    assert dialog.process_combo.currentText() == '404'
    assert dialog.state_lineedit.text() == 'Finished'
    assert dialog.dir_lineedit.text() == '/some/other/dir'
    assert dialog.interpreter_lineedit.text() == '/spam/interpreter'
    assert (dialog.log_textedit.toPlainText()
            == 'Terminated for some reason...\n')



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
#

"""Tests for servermanager.py"""

# Standard library imports
import datetime
import os.path as osp

# Third party imports
import pytest

# Qt imports
from qtpy.QtCore import QByteArray, QProcess, QTimer

# Local imports
from spyder_notebook.utils.servermanager import (
    ServerManager, ServerProcess, ServerState)


@pytest.mark.parametrize('start_arg', [True, False])
def test_get_server_without_servers(mocker, start_arg):
    """Test that .get_server() calls .start_server() when there are no servers
    only if `start` is True."""
    serverManager = ServerManager()
    mock_start = mocker.patch.object(serverManager, 'start_server')
    filename = osp.abspath('ham.ipynb')
    interpreter = '/ham/imterpreter'

    res = serverManager.get_server(filename, interpreter, start=start_arg)

    assert res is None
    if start_arg:
        mock_start.assert_called_once_with(filename, interpreter)
    else:
        mock_start.assert_not_called()


@pytest.mark.parametrize(
    ('nbdir', 'interpreter', 'state',              'result', 'start'),
    [('foo',  'ham',         ServerState.RUNNING,   True,     False),
     ('',     'ham',         ServerState.RUNNING,   True,     False),
     ('foo',  'ham',         ServerState.STARTING,  False,    False),
     ('foo',  'ham',         ServerState.TIMED_OUT, False,    True),
     ('bar',  'ham',         ServerState.RUNNING,   False,    True),
     ('foo',  'cheese',      ServerState.RUNNING,   False,    True)])
def test_get_server_with_server(
        mocker, nbdir, interpreter, state, result, start):
    """Test that .get_server() returns a suitable server if it is accepting
    requests, and that it start up a new server unless a suitable server exists
    that is either starting up or running. Here, a suitable server is a server
    which renders the given notebook with the correct interpreter."""
    serverManager = ServerManager()
    mock_start = mocker.patch.object(serverManager, 'start_server')
    filename = osp.abspath('foo/ham.ipynb')
    server_info = mocker.Mock(spec=dict)
    server = ServerProcess(
        mocker.Mock(spec=QProcess), osp.abspath(nbdir), interpreter,
        state=state, server_info=server_info)
    serverManager.servers.append(server)

    res = serverManager.get_server(filename, interpreter='ham')

    if result:
        assert res == server_info
    else:
        assert res is None
    if start:
        mock_start.assert_called_once_with(filename, 'ham')
    else:
        mock_start.assert_not_called()


@pytest.mark.parametrize(('dark', 'under_home'),
                         [(True, True), (False, True), (False, False)])
def test_start_server(mocker, dark, under_home):
    """Test that .start_server() starts a process with the correct script file,
    notebook dir, and dark argument; that it stores the server process in
    `.servers`; and that it calls ._check_server_running()."""
    serverManager = ServerManager(dark)
    mock_check = mocker.patch.object(serverManager, '_check_server_started')
    mock_QProcess = mocker.patch(
        'spyder_notebook.utils.servermanager.QProcess', spec=QProcess)
    mocker.patch(
        'spyder_notebook.utils.servermanager.get_home_dir', return_value='foo')
    if under_home:
        filename = osp.join('foo', 'bar', 'ham.ipynb')
        nbdir = osp.join('foo')
    else:
        filename = osp.join('notfoo', 'bar', 'ham.ipynb')
        nbdir = osp.join('notfoo', 'bar')

    serverManager.start_server(filename, '/ham/interpreter')

    mock_QProcess.return_value.start.assert_called_once()
    args = mock_QProcess.return_value.start.call_args[0]
    import spyder_notebook.server.main
    assert args[1][0] == spyder_notebook.server.main.__file__
    assert '--notebook-dir={}'.format(nbdir) in args[1]
    assert ('--dark' in args[1]) == dark
    assert len(serverManager.servers) == 1
    assert serverManager.servers[0].process == mock_QProcess.return_value
    assert serverManager.servers[0].notebook_dir == nbdir
    mock_check.assert_called_once()


def test_check_server_started_if_started(mocker, qtbot):
    """Test that .check_server_started() emits sig_server_started if there
    is a json file with the correct name and completes the server info."""
    fake_open = mocker.patch('spyder_notebook.utils.servermanager.open',
                             mocker.mock_open(read_data='{"foo": 42}'))
    mocker.patch('spyder_notebook.utils.servermanager.jupyter_runtime_dir',
                 return_value='runtimedir')
    mock_process = mocker.Mock(spec=QProcess, processId=lambda: 7)
    server_process = ServerProcess(mock_process, 'notebookdir', 'interpreter')
    serverManager = ServerManager()

    with qtbot.waitSignal(serverManager.sig_server_started):
        serverManager._check_server_started(server_process)

    fake_open.assert_called_once_with(
        osp.join('runtimedir', 'nbserver-7.json'), encoding='utf-8')
    assert server_process.state == ServerState.RUNNING
    assert server_process.server_info == {'foo': 42}


def test_check_server_started_if_not_started(mocker, qtbot):
    """Test that .check_server_started() repeats itself on a timer if there
    is no json file with the correct name."""
    fake_open = mocker.patch('spyder_notebook.utils.servermanager.open',
                             side_effect=OSError)
    mocker.patch('spyder_notebook.utils.servermanager.jupyter_runtime_dir',
                 return_value='runtimedir')
    mock_QTimer = mocker.patch('spyder_notebook.utils.servermanager.QTimer',
                               spec=QTimer)
    mock_process = mocker.Mock(spec=QProcess, processId=lambda: 7)
    server_process = ServerProcess(mock_process, 'notebookdir', 'interpreter')
    serverManager = ServerManager()

    serverManager._check_server_started(server_process)

    fake_open.assert_called_once_with(
        osp.join('runtimedir', 'nbserver-7.json'), encoding='utf-8')
    assert server_process.state == ServerState.STARTING
    mock_QTimer.singleShot.assert_called_once()


def test_check_server_started_if_timed_out(mocker, qtbot):
    """Test that .check_server_started() emits sig_server_timed_out if after
    an hour there is still no json file."""
    fake_open = mocker.patch('spyder_notebook.utils.servermanager.open',
                             side_effect=OSError)
    mocker.patch('spyder_notebook.utils.servermanager.jupyter_runtime_dir',
                 return_value='runtimedir')
    mock_process = mocker.Mock(spec=QProcess, processId=lambda: 7)
    one_hour_ago = datetime.datetime.now() - datetime.timedelta(hours=1)
    server_process = ServerProcess(
        mock_process, 'notebookdir', 'interpreter', starttime=one_hour_ago)
    serverManager = ServerManager()

    with qtbot.waitSignal(serverManager.sig_server_timed_out):
        serverManager._check_server_started(server_process)

    fake_open.assert_called_once_with(
        osp.join('runtimedir', 'nbserver-7.json'), encoding='utf-8')
    assert server_process.state == ServerState.TIMED_OUT


def test_check_server_started_if_errored(mocker, qtbot):
    """Test that .check_server_started() does not do anything if server state
    is ERROR."""
    fake_open = mocker.patch('spyder_notebook.utils.servermanager.open')
    mock_QTimer = mocker.patch('spyder_notebook.utils.servermanager.QTimer',
                               spec=QTimer)
    mock_process = mocker.Mock(spec=QProcess, processId=lambda: 7)
    server_process = ServerProcess(
        mock_process, 'notebookdir', 'interpreter', state=ServerState.ERROR)
    serverManager = ServerManager()

    serverManager._check_server_started(server_process)

    fake_open.assert_not_called()
    mock_QTimer.assert_not_called()
    assert server_process.state == ServerState.ERROR


def test_shutdown_all_servers(mocker):
    """Test that .shutdown_all_servers() does shutdown all running servers,
    but not servers in another state."""
    mock_shutdown = mocker.patch(
        'spyder_notebook.utils.servermanager.notebookapp.shutdown_server')
    server1 = ServerProcess(
        mocker.Mock(spec=QProcess), '', '', state=ServerState.RUNNING,
        server_info=mocker.Mock(dict))
    server2 = ServerProcess(
        mocker.Mock(spec=QProcess), '', '', state=ServerState.ERROR,
        server_info=mocker.Mock(dict))
    serverManager = ServerManager()
    serverManager.servers = [server1, server2]

    serverManager.shutdown_all_servers()

    assert mock_shutdown.called_once_with(server1.server_info)
    assert server1.state == ServerState.FINISHED
    assert server2.state == ServerState.ERROR


def test_read_standard_output(mocker):
    """Test that .read_standard_output() stores the output."""
    before = 'before\n'
    output = 'Αθήνα\n'  # check that we can handle non-ascii
    mock_read = mocker.Mock(return_value=QByteArray(output.encode()))
    mock_process = mocker.Mock(spec=QProcess, readAllStandardOutput=mock_read)
    server = ServerProcess(mock_process, '', '', output=before)
    serverManager = ServerManager()
    serverManager.servers = [server]

    serverManager.read_server_output(server)

    mock_read.assert_called_once()
    assert server.output == before + output


def test_handle_error(mocker, qtbot):
    """Test that .handle_error() changes the state and emits signal."""
    server = ServerProcess(mocker.Mock(spec=QProcess), '', '')
    serverManager = ServerManager()

    with qtbot.waitSignal(serverManager.sig_server_errored):
        serverManager.handle_error(server, mocker.Mock())

    assert server.state == ServerState.ERROR


def test_handle_finished(mocker, qtbot):
    """Test that .handle_finished() changes the state."""
    server = ServerProcess(mocker.Mock(spec=QProcess), '', '')
    serverManager = ServerManager()

    serverManager.handle_finished(server, mocker.Mock(), mocker.Mock())

    assert server.state == ServerState.FINISHED
